"""
Internal message dispatcher
"""

# 	Copyright (c) 2005-2007 ALT Linux, Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 2 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


from threading import Thread
from Queue import Queue
from cxutil.UID import UID
import types


def skip(*opts):
	pass

def dump(packet):
	print " ---- "
	print "\tto: %s" % (packet["to"])
	print "\tfrom: %s" % (packet["from"])
	print "\tdata: ",
	print packet["data"]


class connector(object):
	"""
	Two-way thread-safe bus connector for message delivering between two objects.
	It can not be used for any dispatcher-specific operations like subscribing etc, only
	for message transporting.
	
	Attributes:
		* rx (Queue) -- receive queue
		* tx (Queue) -- transmit queue
		* address (UID) -- unical identifier that is used by dispatcher2 to find a recipient
	
	Usually, you don't need to create an instances of this class by hands, it is automatically
	created by dispatcher2.connect()
	
	TODO:
	* create close() and check() methods
	* use weak references to already created Queue() instances
	* define and raise own exceptions
	"""
	
	rx = None
	tx = None
	address = None
	args = None
	
	def __init__(self,wires=(None,None)):
		'''
		Creates new connector for virtual "path cord"  with two wires ("full duplex" :)
		
		It takes a tuple as argument. The first member is a "tx" wire, the second -- "rx".
		If rx or tx "wire" is None, then it will be created.
		'''
		if wires[0]: self.tx = wires[0]
		else: self.tx = Queue()
		if wires[1]: self.rx = wires[1]
		else: self.rx = Queue()
		self.address = UID()

	def pre(self,args):
		'''
		Pre-input hook
		'''
		pass

	def cross(self):
		'''
		Returns a tuple of two wires that can be used to create the second connector
		for this "path cord". Note, that wires are crossed over in the tuple.
		'''
		return (self.rx,self.tx)

	def empty(self):
		'''
		Returns True, if there are no packets in rx
		'''
		return self.rx.empty()

	def drop(self):
		'''
		Drop packets from rx queue
		'''
		try:
			while True:
				self.rx.get_nowait()
		except:
			pass

	def put(self,address,message):
		'''
		Create a packet with "to" and "from" address fields and put it into the tx wire.
		Parameters:
			address (string) -- an address of a recipient
			message (any) -- a message that must be delivered
		
		If address is None, the message must be transported as is (connector assumes
		that it is already a packet)
		'''
		self.pre(self.args)

		if address:
			packet = {}
			packet["to"] = address
			packet["from"] = self.address
			packet["data"] = message
		else:
			packet = message
		result = self.tx.put(packet)
		return result

	def get(self,addr={},timeout=None):
		'''
		Get a packet from rx wire, strip data from it and return
		'''
		try:
			packet = self.rx.get(True,timeout)
		except:
			return None
		addr["to"] = packet["to"]
		addr["from"] = packet["from"]
		return packet["data"]
	
	def put2(self,address,message,addr={}):
		'''
		Put a message in tx and the get from rx
		'''
		self.put(address,message)
		return self.get(addr)

	def get_nowait(self,addr={}):
		'''
		Non-blocking get from rx wire
		'''
		try:
			packet = self.rx.get_nowait()
			addr["to"] = packet["to"]
			addr["from"] = packet["from"]
			result = packet["data"]
		except:
			result = False
		
		return result

class dispatcher2(Thread):
	"""
	Multiple-in multiple-out thread-safe message dispatcher
	
	It works like a router engine. It keeps two major dictionaries:
		* __arp_ -- a hash table, where by UID of client's connector you can find local connector
		* __multicast_ -- a dictionary, where by name of multicast group you can find a list of
		  local connectors of subscribed lines
	
	A message, with which the dispatcher2 works, is a dictionary with three fields.
		* "to" -- UID of a recipient or name of a group
		* "from" -- UID of a sender connector
		* "data" -- the data that must be transported
	
	When a message arrives, dispatcher2 looks into "to" field of the message and chooses a group
	or a particular recipient which must receive this message. A group is created when the first
	client calls subscribe() method.
	
	TODO:
	* create and raise own exeptions like "destination unreachable" etc.
	"""

	__input_ = None
	__multicast_ = None
	__arp_ = None
##	FIXME: make locking while any calls!
	__subscribe_lock = None

	def __init__(self,name="dispatcher2"):
		'''
		Creates a new dispatcher2 thread named by name argument
		'''
		Thread.__init__(self)
		self.setName(name)
		self.__input_ = Queue()
		self.__arp_ = {}
		self.__multicast_ = {}
		self.hook = skip

	def connect(self):
		"""
		Returns a client's connector for new path cord
		"""
##		create the local connector
		local = connector((None,self.__input_))
##		cross-over client's connector
		client = connector(local.cross())
##		register local for client's uid
		self.__arp_[client.address] = local
		return client

	def disconnect(self,address):
		'''
		Remove a client from arp
		'''
		if self.__arp_.has_key(address):
			del self.__arp_[address]
			return True
		else:
			return False
	
	def subscribe(self,address,group):
		'''
		Subscribes an address to a multicast group, creates a group if there is no
		such key
		'''
		if not self.__multicast_.has_key(group):
			self.__multicast_[group] = []
		
		self.__multicast_[group].append(self.__arp_[address])

	def unsubscribe(self,address,group):
		'''
		Remove a subsciption from a multicast group
		'''
		if self.__multicast_.has_key(group):
			if self.__arp_[address] in self.__multicast_[group]:
				self.__multicast_[group].remove(self.__arp_[address])

	def run(self):
		'''
		Main retranslation cycle
		'''
		self.put({"to":"logger","from":"dispatcher2","data":("info","dispatcher thread " + self.getName() + " startup")})
		while True:
##			get a message
			packet = self.__input_.get()
			self.hook(packet)

##			analyse it by itself
			if packet["to"] == self.getName():
				if self.magic(packet["data"],packet["from"]):
					return
				continue
			
			self.put(packet)

	def broadcast(self,message):
##		process broadcast messages
		for i in self.__arp_.keys():
			self.__arp_[i].put(i,message)


	def put(self,packet):
##		search a group
		if self.__multicast_.has_key(packet["to"]):
			for i in self.__multicast_[packet["to"]]:
				i.put(None, packet)
		
##		search a unicast
		if self.__arp_.has_key(packet["to"]):
			self.__arp_[packet["to"]].put(None, packet)
		
	def magic(self,text,addr):
		'''
		Analyses text and tries to interpret it as a "magic" command. Returns True, if the dispatcher2
		instance must break run() cycle and exit thread
		'''

		if text == "shutdown":
			self.put({"to":"logger", "from":"dispatcher2", "data":("info","dispatcher thread " + self.getName() + " shutdown")})
			self.broadcast("shutdown")
			return True

		if text == "connect":
			self.put({"to": addr, "from": "dispatcher2", "data": self.connect()})

		if text == "debug":
			self.hook = dump
			self.put({"to": addr, "from": "dispatcher2", "data": "packet debug has been turned ON"})

		if text == "no debug":
			self.hook = skip
			self.put({"to": addr, "from": "dispatcher2", "data": "packet debug has been turned OFF"})

		if type(text) == types.DictType:
			if text.has_key("disconnect"):
				self.disconnect(text["disconnect"])

			if text.has_key("subscribe"):
				self.subscribe(text["subscribe"],text["group"])

			if text.has_key("hook"):
				# set hook for a connection
				c = self.__arp_[addr]
				c.pre = text["hook"]
				c.args = text["args"]

##		no reasons to break the main cycle found so far, return False...
		return False

	def arp(self):
		return self.__arp_
	
	def groups(self):
		return self.__multicast_

	def subscribers(self,group):
		return self.__multicast_[group]
