"""
This module implements a simple XML-RPC transport over AF_UNIX socket.
The reason is that we can obtain pid/uid/gid of a connected process with
simple getsockopt() call. So, we can delegate user authentication to the OS
level.

An operator can use standard terminal access via ssh or telnet. Our console
client must be used as a login shell. Once it has started, all authorization should
be done with our authorization core.
"""

# 	Copyright (c) 2005-2007 ALT Linux, Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 2 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from socket import SOL_SOCKET
import struct
import xmlrpclib
from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler, SimpleXMLRPCDispatcher
from SocketServer import UnixStreamServer, TCPServer
TCPServer.allow_reuse_address = True

from mangle import mangler
from threading import Thread, Lock
from cxutil.UID import UID

from executive.runner import runner as _runner
import orb.auth

import os
import sys
import pwd
import grp

import types

if sys.platform[:5] == "linux":
	SO_PEERCRED = 17
elif sys.platform[:7] == "freebsd":
	from ctypes.util import find_library
	from ctypes import c_uint32,CDLL,byref
	libc = CDLL(find_library("c"))
	

class DebugXMLRPCRequestHandler(SimpleXMLRPCRequestHandler):
	'''
	Wrapper for SimpleXMLRPCRequestHandler. Gets credentials info
	'''

	def do_POST(self):
		try:
			data = self.rfile.read(int(self.headers["content-length"]))
			response = self.server._marshaled_dispatch(data, getattr(self, '_dispatch', None))

			if self.protodebug:
				print data
				print response

		except:
			self.send_response(500)
			self.end_headers()
			raise
		else:
			# got a valid XML RPC response
			self.send_response(200)
			self.send_header("Content-type", "text/xml")
			self.send_header("Content-length", str(len(response)))
			self.end_headers()
			self.wfile.write(response)
		
			# shut down the connection
			self.wfile.flush()
			self.connection.shutdown(1)

class UnixXMLRPCRequestHandler(SimpleXMLRPCRequestHandler):
	'''
	Wrapper for SimpleXMLRPCRequestHandler. Gets credentials info
	'''

	def do_POST(self):
		try:
##			get peer credentials
			if sys.platform[:5] == "linux":
				(pid,uid,gid) = struct.unpack("III",self.rfile._sock.getsockopt(SOL_SOCKET,SO_PEERCRED,12))
			elif sys.platform[:7] == "freebsd":
				pid = 0
				uid = c_uint32()
				gid = c_uint32()
				a = libc.getpeereid(self.rfile._sock.fileno(), byref(uid), byref(gid))
				uid = uid.value
				gid = gid.value
			else:
				raise

			data = self.rfile.read(int(self.headers["content-length"]))

##			inject credentials into params and repack back into XML-RPC request
##			8<------------------------------------			
			params, method = xmlrpclib.loads(data)
			params = list(params)
			params[0]["pid"] = pid
			params[0]["user"] = pwd.getpwuid(uid).pw_name
			params[0]["group"] = grp.getgrgid(gid).gr_name
			data = xmlrpclib.dumps(tuple(params),method)
##			8<------------------------------------
			

			response = self.server._marshaled_dispatch(data, getattr(self, '_dispatch', None))

			if self.protodebug:
				print data
				print response

		except:
			self.send_response(500)
			self.end_headers()
			raise
		else:
			# got a valid XML RPC response
			self.send_response(200)
			self.send_header("Content-type", "text/xml")
			self.send_header("Content-length", str(len(response)))
			self.end_headers()
			self.wfile.write(response)
		
			# shut down the connection
			self.wfile.flush()
			self.connection.shutdown(1)


class DebugXMLRPCServer(TCPServer, SimpleXMLRPCDispatcher):
	'''
	Simple XML-RPC server, implemented over AF_TCP sockets
	'''
	
	def __init__(self, addr, requestHandler=DebugXMLRPCRequestHandler, logRequests=0, debug=False):
		self.logRequests = logRequests
		SimpleXMLRPCDispatcher.__init__(self)
		requestHandler.protodebug = debug
		TCPServer.__init__(self, addr, requestHandler)

class UnixXMLRPCServer(UnixStreamServer, SimpleXMLRPCDispatcher):
	'''
	Simple XML-RPC server, implemented over AF_UNIX sockets
	'''
	
	def __init__(self, addr, requestHandler=UnixXMLRPCRequestHandler, logRequests=0, trusted=False, debug=False):
		self.logRequests = logRequests
		# FIXME: check version
		if sys.version_info[1] == 4:
			# python 2.4
			SimpleXMLRPCDispatcher.__init__(self)
		elif sys.version_info[1] == 5:
			# python 2.5
			SimpleXMLRPCDispatcher.__init__(self,allow_none = False, encoding = "utf-8")
		else:
			# any other version
			raise Exception("Unsupported python version")

		if trusted:
			requestHandler = DebugXMLRPCRequestHandler
		requestHandler.protodebug = debug
		UnixStreamServer.__init__(self, addr, requestHandler)


class register(object):
	'''
	Registration manager
	'''
	bus = None

	def __init__(self,bus,prefixes):
		self.bus = bus

	def register_service(self,credentials,sid,path):
		'''
		Register an external service
		'''
		return self.bus.put2("orb",
			{
				"map": "register_service",
				"id": sid,
				"path": path
			}
		)

	def register_module(self,credentials,sid,path):
		'''
		Register a module
		'''
		xInstance = orb.auth.proxy(path)
		if xInstance:
			return self.bus.put2("orb",
				{
					"map": "register_module",
					"sid": sid,
					"xClass": orb.auth.proxy,
					"xInstance": xInstance
				}
			)
		else:
			return False

	def register_acl(self,credentials,path,mode,name):
		'''
		Register ACL entry
		'''
		return self.bus.put2("orb",
			{
				"map": "register_acl",
				"mode": mode,
				"name": name,
				"path": path,
			}
		)

	def unregister_acl(self,credentials,path,mode,name):
		'''
		Unregister ACL entry
		'''
		return self.bus.put2("orb",
			{
				"map": "unregister_acl",
				"mode": mode,
				"name": name,
				"path": path,
			}
		)


class proxy(object):
	'''
	A session manager that creates and maintains mangler instances while a session is running.
	'''
	sessions = None
	bus = None
	lock = None
	prefixes = None
	
	def __init__(self,bus,prefixes):
		self.sessions = {}
		self.bus = bus
		self.lock = Lock()
		self.prefixes = prefixes
	
	def error(self,report):
		return {"error": report}
	
	def check(self,credentials,sid):
		user = credentials["user"]

		if self.sessions.has_key(user):
			if self.sessions[user].has_key(sid):
				return True

		return False

	def init(self,credentials,opts = {}):
		return {
			"prompt": self.prefixes[0]
		}

	def start(self,credentials):
		self.lock.acquire()
		oid = UID()
		
		mangler_bus = self.bus.put2("dispatcher2","connect")
		runner_bus = self.bus.put2("dispatcher2","connect")
		
		runner = _runner(runner_bus,"runner for " + str(oid))
		runner.start()
		
		user = credentials["user"]
		
		if not self.sessions.has_key(user):
			self.sessions[user] = {}

		if not self.sessions[user].has_key(oid):
			self.sessions[user][oid] = {}

		self.sessions[user][oid]["mangler"] = mangler(mangler_bus,runner.bus.address,oid,self.prefixes)
		self.sessions[user][oid]["runner"] = runner

		self.lock.release()
		return oid


	def parse(self,credentials,cookie,sid,commands,noout=False):

		if not self.check(credentials,sid):
			return self.error("no such session")

		user = credentials["user"]

		for i in commands:
			self.lock.acquire()
			try:
				self.sessions[user][sid]["mangler"].queue += 1
				result = self.sessions[user][sid]["mangler"].parse(credentials,cookie,i[0].encode('utf-8'),i[1],noout)
			except Exception,e:
				import traceback
				print e
				traceback.print_exc()
			self.lock.release()

		return result

	def interrupt(self,credentials,sid,code):
		if not self.check(credentials,sid):
			return self.error("no such session")

		user = credentials["user"]

		self.lock.acquire()
		result = self.sessions[user][sid]["mangler"].interrupt(credentials,code)
		self.lock.release()
		return result

	def input(self,credentials,sid,text):
		if not self.check(credentials,sid):
			return self.error("no such session")

		user = credentials["user"]

		self.lock.acquire()
		try:
			result = self.sessions[user][sid]["mangler"].input(credentials,text.encode('utf-8'))
		except Exception,e:
			import traceback
			print e
			traceback.print_exc()
		self.lock.release()
		return result

	def list(self,credentials,cookie,sid,prefix=""):
		if not self.check(credentials,sid):
			return self.error("no such session")

		user = credentials["user"]
		result = [cookie,]

		self.lock.acquire()
		try:
			result.extend(self.sessions[user][sid]["mangler"].list(credentials,prefix))
		except Exception,e:
			import traceback
			print e
			traceback.print_exc()
		self.lock.release()
		return result
	
	def terminate(self,credentials,sid):
		if not self.check(credentials,sid):
			return False

		user = credentials["user"]

		self.lock.acquire()

		self.bus.put(self.sessions[user][sid]["runner"].bus.address,"shutdown")
		self.bus.put("dispatcher2",{"disconnect": self.sessions[user][sid]["runner"].bus.address})
		self.bus.put("dispatcher2",{"disconnect": self.sessions[user][sid]["runner"].orb_bus.address})
		self.bus.put("dispatcher2",{"disconnect": self.sessions[user][sid]["mangler"].bus.address})
		self.bus.put("dispatcher2",{"disconnect": self.sessions[user][sid]["mangler"].busInstance.address})
		self.bus.put("dispatcher2",{"disconnect": self.sessions[user][sid]["mangler"].busRunner.address})
		self.bus.put("dispatcher2",{"disconnect": self.sessions[user][sid]["mangler"].busState.address})
		
		os.close(self.sessions[user][sid]["mangler"].pipe_read)
		os.close(self.sessions[user][sid]["mangler"].pipe_write)
		os.close(self.sessions[user][sid]["mangler"].ctl_read)
		os.close(self.sessions[user][sid]["mangler"].ctl_write)
		
		del self.sessions[user][sid]
		self.lock.release()
		return True
	

class server(Thread):
	'''
	XML-RPC publisher
	'''
	flags = None
	server = None
	bus = None

	def __init__(self,bus,publish,name="XML-RPC server",socket="socket",trusted=False,debug=False,prefixes=[]):
		Thread.__init__(self)
		self.setName(name)
		self.flags = {}
		self.bus = bus
		
		try:
			if type(socket) == types.TupleType:
				self.server = DebugXMLRPCServer(socket, debug = debug)
			else:
				self.server = UnixXMLRPCServer(socket, trusted = trusted, debug = debug)
			self.server.register_introspection_functions()
			self.server.register_instance(publish(self.bus,prefixes))
		except:
			raise

	
	def run(self):
		while True:
			self.server.handle_request()
			if self.bus.get_nowait() == "shutdown":
				return

